/*
 * Fujitsu lifebook application panel driver
 * apanel.c - the apanel driver
 *
 * (c) Copyright by Willem van Engen
 * See COPYRIGHT for copyright
 * See CREDITS for credits
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/fcntl.h>

#include <dev/smbus/smbconf.h>
#include <dev/smbus/smbus.h>
#include <machine/smb.h>

#include <vm/vm.h>
#include <vm/pmap.h>
#include <machine/md_var.h>
#include <machine/pc/bios.h>

#include "smbus_if.h"
#include "apanel.h"

/* Devicename in debugging info */
#define DEVTITLE APANEL_DEVNAME"0"

/* Devices in the bios */
#define DEV_APPLBTN	1
#define DEV_CDBTN	2
#define DEV_LCD		3
#define DEV_LED		4
/* Number of devices supported */
#define MAXDEVICES	4

/* ascii representations of device names */
const char *device_names[5] = { "", "APPBTN", "CDBTN", "LCD", "LED" };

/* Structure containing data about a specific apanel device */
struct device_data {
  unsigned char exists;		/* =1 if device is detected in system, else 0 */
  unsigned char method;
  unsigned char device;
};
/* Relevant data from bios comes in here */
struct device_data devdata[MAXDEVICES+1];

/* The driver softc structure */
struct apanel_softc {
  dev_t	dev;
  int lcdval;
  int ledval;
  device_t smbus;
};

static int apanel_probe(device_t);
static int apanel_attach(device_t);
static int apanel_detach(device_t);
static void apanel_identify(driver_t*,device_t);
static int apanel_suspend(device_t);
static int apanel_resume(device_t);

static int do_apanel_cmd(struct apanelcmd*,struct apanel_softc*,int);
static int apanel_smb_write(int,int,int,struct apanel_softc*,int);
static int apanel_smb_read(int,int,int*,struct apanel_softc*,int);

static devclass_t apanel_devclass;

static device_method_t apanel_methods[] = {
	/* device interface */
	DEVMETHOD(device_probe,		apanel_probe),
	DEVMETHOD(device_attach,	apanel_attach),
	DEVMETHOD(device_detach,	apanel_detach),
	DEVMETHOD(device_identify,	apanel_identify),
	/* Suspend and resume */
	DEVMETHOD(device_suspend,	apanel_suspend),
	DEVMETHOD(device_resume,	apanel_resume),
	/* smbus interface */
	DEVMETHOD(smbus_intr,		smbus_generic_intr),
	{ 0, 0 }
};

static driver_t apanel_driver = {
	APANEL_DEVNAME,
	apanel_methods,
	sizeof(struct apanel_softc),
};

static	d_open_t	apanelopen;
static	d_close_t	apanelclose;
static	d_write_t	apanelwrite;
static	d_read_t	apanelread;
static	d_ioctl_t	apanelioctl;

static struct cdevsw apanel_cdevsw = {
	/* open */	apanelopen,
	/* close */	apanelclose,
	/* read */	apanelread,
	/* write */	apanelwrite,
	/* ioctl */	apanelioctl,
	/* poll */	nopoll,
	/* mmap */	nommap,
	/* strategy */	nostrategy,
	/* name */	APANEL_DEVNAME,
	/* maj */	APANEL_MAJOR,
	/* dump */	nodump,
	/* psize */	nopsize,
	/* flags */	0,
	/* bmaj */	-1
};

/*
 * Probe function
 *
 * Is the device here, can I talk to it and do I have all the
 * resources I need?
 */
static int
apanel_probe(device_t dev)
{
	u_int32_t fjptr;
	unsigned char i, dno;
	char therearedevs=0;

#ifdef APANEL_DEBUG
	printf("%s: apanel_probe called\n",DEVTITLE);
#endif

	/* Reset values first */
	for (i=0; i<MAXDEVICES; i++) devdata[i].exists=0;

	/* Extract info from bios */
	if ( (fjptr=bios_sigsearch(0,"FJKEYINF",8,0x10,0))==NULL) {
		printf("%s: no Fujitsu lifebook; couldn't find \"FJKEYINF\" in bios.\n",
				DEVTITLE);
		return (-1);
	}
	fjptr=BIOS_PADDRTOVADDR(fjptr)+8;	/* Start after FJKEYINF string */
	for (i=0;i<6;i++,fjptr+=4) {		/* Loop for each device param block */
		dno=*((char*)(fjptr));
		if (dno==0) break;				/* End of block (I guess) */
		if (dno<=MAXDEVICES) {
			devdata[dno].method=*((char*)(fjptr+1));
			/* Currently only smbus supported, not so-called model X */
			if (devdata[dno].method!=1 && devdata[dno].method!=2) {
				printf("%s: incompatible access method for %s\n",
						DEVTITLE, device_names[dno]);
			} else {
				devdata[dno].exists=1;
				therearedevs=1;
				devdata[dno].device=*((char*)(fjptr+3));
#ifndef APANEL_DEBUG
				if (bootverbose)
#endif
					printf("%s: found device %s, slave address = 0x%x\n",
						DEVTITLE,device_names[dno],devdata[dno].device);
			}
		}
	}

	if (therearedevs==0)
		printf("%s: no devices found\n",DEVTITLE);

	device_set_desc(dev, "Fujitsu Lifebook Application Panel");

	return (0);
}

/*
 * apanel_attach
 *
 * Attach device and init
 */
static int
apanel_attach(device_t dev)
{
	struct apanel_softc *sc = (struct apanel_softc *)device_get_softc(dev);
	struct apanelcmd apc;
	unsigned char i;

#ifdef APANEL_DEBUG
	printf("apanel_attach called\n");
#endif

	/* Get device_t of smbus for bus alloc later */
	sc->smbus=device_get_parent(dev);
	sc->dev=make_dev(&apanel_cdevsw, device_get_unit(dev),
			UID_ROOT, GID_WHEEL, 0600, APANEL_DEVNAME);

	/* And finally reset all existing devices */
	for (i=0; i<MAXDEVICES; i++) {
		if (devdata[i].exists==1) {
			apc.device = i;
			apc.cmd = APANEL_CMD_RESET;
			do_apanel_cmd(&apc, sc, 0);
		}
	}

	return (0);
}

/*
 * apanel_detach
 *
 * Clean up everything
 */
static int
apanel_detach(device_t dev)
{
	struct apanel_softc *sc = (struct apanel_softc *)device_get_softc(dev);

#ifdef APANEL_DEBUG
	printf("apanel_detach called\n");
#endif
	destroy_dev(sc->dev);
	return(0);
}

/*
 * apanel_identify
 *
 * Os is enumerating us
 */
static void
apanel_identify(driver_t *driver, device_t parent)
{
	devclass_t dc;
	device_t child;

#ifdef APANEL_DEBUG
	printf("%s: apanel_identify called\n",DEVTITLE);
#endif
	dc = devclass_find(APANEL_DEVNAME);
	if (devclass_get_device(dc, 0)==NULL) {
#ifdef SMBUS_WORKAROUND
		child = BUS_ADD_CHILD(parent, 0, APANEL_DEVNAME, -1);
#else
#		warning "You might want to define SMBUS_WORKAROUND to work around some broken system management bus code."
		child = device_add_child(parent, APANEL_DEVNAME, -1);
#endif
	}
}

static int
apanel_suspend(device_t dev)
{
	struct apanel_softc *sc = (struct apanel_softc *)device_get_softc(dev);

	/* Don't exactly know why, but the windriver does it */
	apanel_smb_write(APANEL_DEV_APPBTN,0x05,0x0000,sc,0);
	return (0);
}

static int
apanel_resume(device_t dev)
{
	struct apanel_softc *sc = (struct apanel_softc *)device_get_softc(dev);

	/* Don't exactly know why, but the windriver does it */
	apanel_smb_write(APANEL_DEV_APPBTN,0x05,0x00ff,sc,0);
	return (0);
}

static int
apanelopen (dev_t dev, int flags, int fmt, struct proc *p)
{
	return (0);
}

static int
apanelclose(dev_t dev, int flags, int fmt, struct proc *p)
{
	return (0);
}

static int
apanelwrite(dev_t dev, struct uio * uio, int ioflag)
{
	return (EINVAL);	/* not supported */
}

static int
apanelread(dev_t dev, struct uio * uio, int ioflag)
{
	/* suggestion:
	 *  - output textual info about driver */
	return (EINVAL);	/* not supported */
}

/*
 * apanelioctl
 *
 * Handles all ioctl's. That's were in this driver the
 * actual work is done.
 */
static int
apanelioctl(dev_t dev, u_long cmd, caddr_t data, int flags, struct proc *p)
{
	struct apanel_softc *sc = 
		(struct apanel_softc *)devclass_get_softc(apanel_devclass, 0);
	struct apanelcmd *apc = (struct apanelcmd *)data;
	int error = 0;

	if (!sc)
		return (EINVAL);

	switch(cmd) {
	case IOCTL_APANEL_CMD:		/* Real command, passed on to do_apanel_cmd */
#ifdef APANEL_DEBUG
		printf("%s: APANEL_CMD(device=0x%x,cmd=0x%x,data=0x%x)\n",
				DEVTITLE,apc->device,apc->cmd,apc->data);
#endif
		error=do_apanel_cmd(apc,sc,flags);
		break;

	case IOCTL_APANEL_CAP_GEN:	/* General capabilities (which devices) */
		/* Existing devices */
		*(int*)data = (devdata[1].exists<<0) + (devdata[2].exists<<1) +
					(devdata[3].exists<<2) + (devdata[4].exists<<3);
#ifdef APANEL_DEBUG
		printf("%s: IOCTL_APANEL_CAP_GEN returning 0x%x\n",
				DEVTITLE,*(int*)data);
#endif
		break;

	case IOCTL_APANEL_VER:		/* Returns driver interface version */
		*(int*)data = 0x100;			/* Version 1.00 */
#ifdef APANEL_DEBUG
		printf("%s: IOCTL_APANEL_VER returning 0x%x\n",DEVTITLE,*(int*)data);
#endif
		break;

	default:					/* Anything else returns error */
		error = ENODEV;
	}

	return (error);
}

/*
 * do_apanel_cmd
 *
 * Executes an apanel command
 */
static int
do_apanel_cmd(struct apanelcmd* apc,struct apanel_softc *sc,int flags)
{
	int error=0;

	switch(apc->cmd) {
	case APANEL_CMD_CAPS:		/* Return device's capabilities */
		switch(apc->device) {
		case APANEL_DEV_LED:
			apc->data = APANEL_CAP_ONOFF | APANEL_CAP_BLINK | 
						APANEL_CAP_RESET | APANEL_CAP_DUTY_FIXED;
			break;

		case APANEL_DEV_LCD:
			apc->data = APANEL_CAP_ONOFF | APANEL_CAP_GET | APANEL_CAP_SET |
						APANEL_CAP_MAX | APANEL_CAP_MIN | APANEL_CAP_RESET;
			break;

		case APANEL_DEV_APPBTN:
			apc->data = APANEL_CAP_GET | APANEL_CAP_RESET;
			break;

		case APANEL_DEV_CDBTN:
			apc->data = APANEL_CAP_GET | APANEL_CAP_RESET;
			break;

		default:
			error=ENODEV;
		}
		break;

	case APANEL_CMD_ONOFF:	/* Turn device on or off */
		if (apc->device==APANEL_DEV_LED) {
			if (apc->data==APANEL_ONOFF_ON) {
				/* led on: write old values to led */
				error=apanel_smb_write(APANEL_DEV_LED,0x10,0x8000,sc,flags);
			} else if (apc->data==APANEL_ONOFF_OFF) {
				/* led off */
				error=apanel_smb_write(APANEL_DEV_LED,0x10,0x0000,sc,flags);
			} else error=ENODEV;
		} else if (apc->device==APANEL_DEV_LCD) {
			if (apc->data==APANEL_ONOFF_ON) {
				/* lcd on again: Write old value to lcd */
				error=apanel_smb_write(APANEL_DEV_LCD,0x06,sc->lcdval,sc,flags);
			} else if (apc->data==APANEL_ONOFF_OFF) {
				/* lcd off: Write zero to lcd */
				error=apanel_smb_write(APANEL_DEV_LCD,0x06,0,sc,flags);
			} else error=ENODEV;
		} else error=ENODEV;
		break;

	case APANEL_CMD_SET:	/* Set (raw) value */
		if (apc->device==APANEL_DEV_LCD) {
			if (apc->data>0 && apc->data<=100) {
				sc->lcdval=apc->data<<8;		/* Save value for later */
				error=apanel_smb_write(APANEL_DEV_LCD,0x6,sc->lcdval,sc,flags);
			} else error=ENODEV;
		} else error=ENODEV;
		break;

	case APANEL_CMD_GET:	/* Get (raw) value */
		switch (apc->device) {
		case APANEL_DEV_LCD:
			error=apanel_smb_read(APANEL_DEV_LCD,0x6,&apc->data,sc,flags);
			apc->data=apc->data>>8;
			break;

		case APANEL_DEV_APPBTN:
			error=apanel_smb_read(APANEL_DEV_APPBTN,0x0,&apc->data,sc,flags);
			apc->data= (~apc->data>>8)&0x0f;	/* Inverted and in highbyte */
			break;

		case APANEL_DEV_CDBTN:
			error=apanel_smb_read(APANEL_DEV_CDBTN,0x1,&apc->data,sc,flags);
			if (!error)
				error=apanel_smb_write(APANEL_DEV_CDBTN,0x1,0,sc,flags);
			break;

		default:
			error=ENODEV;
		}
		break;

	case APANEL_CMD_BLINK:		/* Set blink time */
		if (apc->device==APANEL_DEV_LED) {
#ifdef APANEL_DEBUG
			printf("%s: APANEL_CMD_BLINK(0x%x)\n",DEVTITLE,apc->data);
#endif
			if (apc->data<0x7f) {
				error=apanel_smb_write(APANEL_DEV_LED,0x11,
								(apc->data&0xff)<<8,sc,flags);
			} else error=ENODEV;
		} else error=ENODEV;
		break;

	case APANEL_CMD_DUTY:		/* Set duty cycle */
		if (apc->device==APANEL_DEV_LED) {
#ifdef APANEL_DEBUG
			printf("%s: APANEL_CMD_DUTY(0x%x)\n",DEVTITLE,apc->data);
#endif
			if (apc->data==APANEL_DUTY_EQUAL) {
				error=apanel_smb_write(APANEL_DEV_LED,0x0f,0x00,sc,flags);
			} else if (apc->data==APANEL_DUTY_OTHER1) {
				error=apanel_smb_write(APANEL_DEV_LED,0x0f,0x80,sc,flags);
			} else error=ENODEV;
		} else error=ENODEV;
		break;

	case APANEL_CMD_RESET:		/* Resets device to default */
		if (apc->device==APANEL_DEV_LCD || apc->device==APANEL_DEV_LED) {
			struct apanelcmd apc2 =
				{ apc->device, APANEL_CMD_ONOFF, APANEL_ONOFF_OFF };
			do_apanel_cmd(&apc2, sc, flags);
		}				/* Claim to support reset for all devices (no error)*/
		break;

	case APANEL_CMD_MAX:		/* Returns max value for something */
		switch (apc->device) {
		case APANEL_DEV_LCD:
			if (apc->data==APANEL_CMD_SET) apc->data=99;
			else error=ENODEV;
			break;

		case APANEL_DEV_LED:
			if (apc->data==APANEL_CMD_BLINK) apc->data=0x7f;
			else if (apc->data==APANEL_CMD_DUTY) apc->data=APANEL_DUTY_OTHER1;
			else error=ENODEV;
			break;

		default:
			error=ENODEV;
		}
		break;

	case APANEL_CMD_MIN:		/* Returns min value for something */
		switch(apc->device) {
		case APANEL_DEV_LCD:
			apc->data=1;
			break;

		case APANEL_DEV_LED:
			if (apc->data==APANEL_CMD_BLINK) apc->data=0;
			else if (apc->data==APANEL_CMD_DUTY) apc->data=APANEL_DUTY_EQUAL;
			else error=ENODEV;
			break;

		default:
			error=ENODEV;
		}
		break;

	default:
		error=ENODEV;
	}

	return (error);
}

/*
 * apanel_smb_write
 *
 * Executes a system management bus write for apanel devices
 */
static int
apanel_smb_write(int dev, int func, int value, struct apanel_softc *sc,int fl)
{
	int error;
	device_t apaneldev = devclass_get_device(apanel_devclass,0);

	/* allocate the bus */
	if ((error = smbus_request_bus(sc->smbus, apaneldev,
		(fl & O_NONBLOCK) ? SMB_DONTWAIT : (SMB_WAIT | SMB_INTR))))
		return (error);

	/* Write stuff to smbus */
#ifdef APANEL_DEBUG
	printf("%s: smb_write(dev=0x%x,func=0x%x,value=0x%x)\n",
			DEVTITLE,devdata[dev].device,func,value);
#endif
	error=smbus_writew(sc->smbus,devdata[dev].device,func,value);

	/* release the bus */
	smbus_release_bus(sc->smbus, apaneldev);

	return (error);
}

/*
 * apanel_smb_read
 *
 * Executes a system management bus read for apanel devices
 */
static int
apanel_smb_read(int dev, int func, int *value, struct apanel_softc *sc,int fl)
{
	short val;
	int error;
	device_t apaneldev = devclass_get_device(apanel_devclass,0);

	/* allocate the bus */
	if ((error = smbus_request_bus(sc->smbus, apaneldev,
		(fl & O_NONBLOCK) ? SMB_DONTWAIT : (SMB_WAIT | SMB_INTR))))
		return (error);

	/* Read stuff from smbus */
	error=smbus_readw(sc->smbus,devdata[dev].device,func,&val);
	*value=val;
#ifdef APANEL_DEBUG
	printf("%s: smb_read(dev=0x%x,func=0x%x) = 0x%x\n",
			DEVTITLE,devdata[dev].device,func,val);
#endif

	/* release the bus */
	smbus_release_bus(sc->smbus, apaneldev);

	return error;
}

DRIVER_MODULE(apanel, smbus, apanel_driver, apanel_devclass, 0, 0);

