/*
 * lcdstatd.c
 * 
 * Shows system information on a Fujitsu lifebook lcd
 * application panel driver required.
 * 
 * Copyright by Willem van Engen <wvengen@stack.nl>
 * See COPYRIGHT for copyright
 */
/*
 * TODO:
 *   - add apm handler which sets the display and on resume (or move to driver?)
 *   - use a better way to periodically update the display 
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <machine/apm_bios.h>
#include <sys/syscall.h>
#include <signal.h>
#include <fcntl.h>
#include "apanel.h"

#define HELPMSG	\
"Shows system information on Fujitsu Lifebook lcd\n"\
"\n"\
"Usage: lcdstatd -s l|b|p [-i interval] [-t 1|5|15] [-h] [-w] [-n num]\n"\
"  -i          Update display each interval milliseconds\n"\
"  -s l|b|p    What to display:\n"\
"              l    System load\n"\
"              b    Remaining battery time in minutes\n"\
"              p    Remaining battery percentage\n"\
"  -t 1|5|15   Period of time of load average (1, 5 or 15 minutes)\n"\
"  -n          Number for renice-ing from 20(low priority) to -20 (default 10)\n"\
"  -h          Show this help\n"

#define WHAT_BPERC	1
#define WHAT_BTIME	2
#define WHAT_SLOAD	3

#define APMDEV		"/dev/apm"
#define PANELDEV	"/dev/apanel"

void reload(int);
void quit(int);

const char panelopenerr[]="Could not open panel device.\n";
const char apmopenerr[]="Could not open apm device.\n";

int apmfd=0,panel=0;
char endit=0;
char *err=NULL;

int main(int argc,char* argv[]) {
  int niceness=10;
  int sampleno=0;
  long int interval=1000;
  int num,oldnum=-1;
  int wrap=0;
  double loads[3];
  char ch;
  struct apm_info apminfo;
  char what=0;
  struct apanelcmd apc;
  int lcdmin=1, lcdmax=99;

#define OPTS "hi:t:ws:hn:"

  /* Parse cmdline */
  while ((ch=getopt(argc,argv,OPTS)) != -1) {
    switch(ch) {
    case 'i':	/* Update interval */
      interval=atoi(optarg)*1000;
      if (interval=0) {
        fprintf(stderr,"An interval of 0 will load your system too heavily\n");
	exit(1);
      }
      break;

    case 't':	/* Period of time of load average */
      switch(atoi(optarg)) {
      case 1:
        sampleno=0;
	interval=1000;
	break;

      case 5:
	sampleno=1;
	interval=5000;
	break;

      case 15:
	sampleno=2;
	interval=15000;
	break;

      default:
	fprintf(stderr,"Valid time periods are 1, 5 and 15 minutes\n");
	exit(1);
      }
      break;

    case 'w':	/* Wrap number */
      wrap=1;
      break;

    case 'h':	/* Help message */
      printf(HELPMSG);
      return 0;

    case 's':	/* What to display */
      if (strlen(optarg)!=1) {
        fprintf(stderr,"Wrong number of arguments to option -s."
                       "Use -h for help.");
	exit(1);
      }
      switch(optarg[0]) {
      case 'b':		/* Display battery time */
        what=WHAT_BTIME;
        break;

      case 'p':		/* Display battery percentage */
        what=WHAT_BPERC;
        break;
	
      case 'l':		/* Display system load */
	what=WHAT_SLOAD;
	break;

      default:
        fprintf(stderr,"Don't know what to display (%c). "
                       "Use -h for help.\n",optarg[0]);
	exit(1);
      }
      break;

    case 'n':		/* Renice value */
      niceness=atoi(optarg);
      break;

    default:
      fprintf(stderr,"Unknown option: %c\n",ch);
      exit(1);
    }
  }

  /* We need options to indicate what to do */
  if (what==0) {
    fprintf(stderr,"You might want to indicate what the lcd should display.\n"
            "Use -h for help.\n");
    exit(1);
  }

  /* Setup signal handling */
  signal(SIGINT,quit);
  signal(SIGHUP,reload);

  /* Open application panel device */
  if ((panel=open(PANELDEV,O_RDONLY))==NULL) {
    fprintf(stderr,"Could not open " PANELDEV " application panel device.\n"
		   "Is this a Fujitsu Lifebook and is apanel.ko loaded?\n");
    exit(1);
  }

  /* Do we neet apm? */
  if (what==WHAT_BTIME || what==WHAT_BPERC) {
    /* check if apm supports it */	  
    if ((apmfd=open(APMDEV,O_RDONLY))==NULL) {
      fprintf(stderr,"Could not open /dev/apm.\n");
      exit(1);
    }
  }

  /* Renice myself */
  setpriority(PRIO_PROCESS,0,niceness);

  if (daemon(NULL,NULL)==-1) {
    fprintf(stderr,"Could not become daemon\n");
    return 1;
  }

  apc.device = APANEL_DEV_LCD;
  apc.cmd = APANEL_CMD_MIN;
  apc.data = APANEL_CMD_SET;
  if (ioctl(panel,IOCTL_APANEL_CMD,&apc)!=-1) lcdmin=apc.data;
  apc.cmd = APANEL_CMD_MAX;
  apc.data = APANEL_CMD_SET;
  if (ioctl(panel,IOCTL_APANEL_CMD,&apc)!=-1) lcdmax=apc.data;

  apc.cmd = APANEL_CMD_SET;
  /* Main daemon loop. Gets each time the requested information and
     puts it on the lcd using the apanel device */
  while (1) {
    if (endit==1) break;
    switch(what) {
    case WHAT_SLOAD:		/* System load */
      if (getloadavg(loads,3)!=-1) {
        num=(int)(100*loads[sampleno]);
      } else {
	continue;
      }
      break;

    case WHAT_BPERC:	/* Percent of battery remaining */
      if (ioctl(apmfd,APMIO_GETINFO,&apminfo)==-1) continue;
      num=apminfo.ai_batt_life;
      break;

    case WHAT_BTIME:		/* Batterytime remaining */
      if (ioctl(apmfd,APMIO_GETINFO,&apminfo)==-1) continue;
      num=apminfo.ai_batt_time/60;
      if (num==0) num=-1;
      break;
    }

    if (num==-1) {
        num=lcdmax+1;
    } else {
        if (num>lcdmax) num=lcdmax;
	else if (num<lcdmin) num=lcdmin;
    }

    if (num!=oldnum) {			/* No need for update if no change */
       apc.data=num;
       if (ioctl(panel,IOCTL_APANEL_CMD,&apc)!=-1) {
         oldnum=num;			/* Save this for comparison later */
       }
    }
    usleep(interval);			/* Sleep for a while */
  }

  /* Close stuff and exit */
  if (apmfd!=0) close(apmfd);
  if (panel!=NULL) close(panel);
  if (err!=NULL) fprintf(stderr,"%s",*err);
  return 0;
}

/* Close all filehandles and reopen them */
void reload(int sig) {
  if (apmfd!=0) close(apmfd);
  if (panel!=NULL) close(panel);
  if ((apmfd=open(APMDEV,O_RDONLY))==0) {
    endit=1;
    err=(char*)&apmopenerr;
  }
  if ((panel=open(PANELDEV,O_RDONLY))==NULL) {
    endit=1;
    err=(char*)&panelopenerr;
  }
}

/* Indicate we want to quit */
void quit(int sig) {
  endit=1;
}

